/***
 **
 *  File: Parser.cpp
 *  Created: Dec 12, 2009 4:05 PM
 *
 *  Author: Xiao Yang <isuxyang@gmail.com>
 *
 *  Copyright 2009 Xiao Yang, Karin Dorman, Srinivas Aluru
 *
 *  This file is part of Reptile.
 *
 *  Reptile is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Reptile is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with Libpnorm. If not, see <http://www.gnu.org/licenses/>.
 *
 */
 

#include "Parser.h"

/*
 * Load short reads and quality information
 */
void Parser::load(const Para& myPara) {

    bIO::FASTA_input fasta_sr(myPara.iFaName);
    bIO::FASTA_input fasta_q(myPara.iQName);

    if (fasta_sr.operator bool() == false) {
        std::cout << "open " << myPara.iFaName << "failed, does it exist? \n";
        exit(1);
    }

    std::cout << "Loading reads (batch size: " << (double) myPara.batchSize / 1000
            << " K) \n\t Batch:";
    int batch = 0;
    double timing = jaz::get_time();
    while (fasta_sr.operator bool()) {

        std::cout << batch << "\t";
        batch++;

        Seq mySeq;
        Kmer spKArray, lpKArray;

        mySeq.retrieve_seqs_from_fa(fasta_sr, myPara.batchSize, false);
        if (myPara.QFlag) {
            mySeq.retrieve_seqs_from_fa(fasta_q, myPara.batchSize, true);
        }

        spKArray.getKmer(mySeq, 0, myPara);
        lpKArray.getKmer(mySeq, myPara.step, myPara);

        merge(snKArray_, spKArray.getElem(), myPara.K);
        merge(lnKArray_, lpKArray.getElem(), myPara.K + myPara.step);
        print_time("", timing);


        /*
        std::cout << "search ";
        size_t ID = 876452;
        //toID (ID, "ggttttcatgtc", 12);
        std::cout << ID << "\t and ";
        int idx1 = bisearch (snKArray_, snKArray_.size(), ID);
        //toID (ID, "gacatgaaaacc", 12);
        //std::cout << ID << "\n";
        //int idx2 = bisearch (snKArray_, snKArray_.size(), ID);
        if (idx1 != -1) {
            std::cout << "found 1\n\n";
            exit(1);
        }
        //if(idx2 != -1){
        //    std::cout << "found 2\n\n";
        //}
        */
         
    }


    //std::cout << "final snKArray size = " << snKArray_.size() << "\n";
    //std::cout << "final lnKArray size = " << lnKArray_.size() << "\n";
    std::cout << "Loading done! Approximately "
            << (double) batch * myPara.batchSize / 1000000 << " Million Reads. \n\n";

    //printKID2Cnt(myPara);
}

int bisearch(const kcvec_t& KArray, int size, uint64_t elem) {
    int lb = 0, ub = size - 1, mid;
    while (lb <= ub) {
        mid = (lb + ub) / 2;
        if (KArray[mid].ID == elem)
            return mid;
        else if (KArray[mid].ID < elem)
            lb = mid + 1;
        else if (KArray[mid].ID > elem)
            ub = mid - 1;
    }
    return -1;
}

void merge(kcvec_t& KArray, const std::vector<kc_s>& KMap, int Klen) {

    int incre = KMap.size();
    int oriSize = (int) KArray.size();
    KArray.reserve(oriSize + incre);
    int trueIncre = 0;

    for (int i = 0; i < incre; ++i) {

        uint64_t ID;
        if (toID(ID, KMap[i].repaddr, Klen)) {

            int idx = bisearch(KArray, oriSize, ID);
            if (idx == -1) {
                KArray.push_back(kc_t(ID, KMap[i].goodCnt, KMap[i].cnt));
                trueIncre++;
            } else {
                KArray[idx].cnt += KMap[i].cnt;
                KArray[idx].goodCnt += KMap[i].goodCnt;
            }
        }
    }
    KArray.resize(oriSize + trueIncre);
    std::sort(KArray.begin(), KArray.end(), Knumcomp());
}

void print(int i) {
    std::cout << " " << i;
}

void printHex(int i) {
    std::cout << " " << std::hex << i;
}

void Parser::tableMaker(const Para& myPara) {

    std::cout << "Constructing Facility Tables ... \n";
    /*
     * 1. split K positions into random chunks based
     *    on myPara.k and myPara.eSearch
     */
    ivec_t indices(myPara.K, 0);
    for (int i = 0; i < myPara.K; ++i, indices[i] = i);
    std::random_shuffle(indices.begin(), indices.end());
    //std::for_each(indices.begin(), indices.end(), print);  std::cout <<"\n";

    int unitSize = (log2(myPara.eSearch) / 2);
    int num = myPara.K / unitSize;
    for (int i = 0; i < num; ++i) {
        ivec_t tmp;
        tmp.insert(tmp.end(), indices.begin() + i*unitSize,
                indices.begin() + (i + 1) * unitSize);
        maskIdx_.push_back(tmp);
    }

    int lastUnitSize = myPara.K % unitSize;
    if (lastUnitSize != 0) {
        ivec_t tmp;
        tmp.insert(tmp.end(), indices.begin() + num*unitSize, indices.end());
        maskIdx_.push_back(tmp);
        num++;
    }

    // debug
    //for (int i = 0; i < num; ++i) {
    //    std::for_each(myChunk[i].begin(), myChunk[i].end(), print);
    //    std::cout << "\n";
    //}

    //std::cout << "\t" << num << " Tables to be created\n";

    /*
     * 2. Create Masks for tables
     */
    masks_.resize(num);
    for (int i = 0; i < num; masks_[i] = 0xFFFFFFFF, i++);
    for (int i = 0; i < num; ++i) {
        for (int j = 0; j < maskIdx_[i].size(); ++j) {
            masks_[i] &= ~(1 << (2 * maskIdx_[i][j]));
            masks_[i] &= ~(1 << (2 * maskIdx_[i][j] + 1));
        }
    }

    //std::for_each(masks_.begin(), masks_.end(), printHex);
    //std::cout << "\n";

    /*
     * 3. Create Tables Then sort according to masks for each table
     */
    int tableSize = snKArray_.size();
    table_.reserve(num * tableSize);
    for (int i = 0; i < tableSize; ++i) {
        table_.push_back((uint32_t) snKArray_[i].ID);
    }
    for (int i = 1; i < num; ++i) {
        table_.insert(table_.end(), table_.begin(), table_.begin() + tableSize);
    }
    /*
     * sort tables
     */
    for (int i = 0; i < num; ++i) {
        std::sort(table_.begin() + i*tableSize,
                table_.begin() + (i + 1) * tableSize, TComp(masks_[i]));
    }

    std::cout << "\tdone !\n\n";

}

void Parser::ec(const Para& myPara) {

    std::cout << "Start Error Correction in batches...(batch size: "
            << (double) myPara.batchSize / 1000 << " K) \n\t Batch:";

    bIO::FASTA_input fasta_sr(myPara.iFaName);
    bIO::FASTA_input fasta_q(myPara.iQName);

    if (fasta_sr.operator bool() == false) {
        std::cout << "open " << myPara.iFaName << "failed :|\n";
        exit(1);
    }

    int batch = 0;
    double timing = jaz::get_time();

    while (fasta_sr.operator bool()) {

        std::cout << batch << "\t";
        batch++;

        Seq mySeq;
        mySeq.retrieve_seqs_from_fa(fasta_sr, myPara.batchSize, false);
        if (myPara.QFlag) {
            mySeq.retrieve_seqs_from_fa(fasta_q, myPara.batchSize, true);
        }

        //---------------------------------------------------------------------
        // now start correcting this batch -- could be multi-threaded
        //---------------------------------------------------------------------
        int numSeq = mySeq.getSL().size();

        for (int i = 0; i < numSeq; ++i) {
            int pos = mySeq.getSL()[i];
            char* addr = const_cast<char*> (&mySeq.getS()[pos]);
            char* qAddr = NULL;
            if (myPara.QFlag) {
                qAddr = const_cast<char*> (&mySeq.getQ()[pos]);
            }
            readEC(addr, qAddr, myPara);
            readID_++;
        }// for
        //---------------------------------------------------------------------                
        print_time("", timing);
    }//while (fasta_sr.operator bool())

    std::cout << "\tdone !\n\n";
}

void Parser::readEC(char* addr, char* qAddr, const Para& myPara) {

//    std::cout << "Read: " << readID_ << "\n"; // debug
 //   std::cout << addr << "\n"; //debug
//    if (readID_ > 1000){
//        output(myPara.oErrName);
//        exit(1);
//    }
    

    //if (readID_ > 1000) {
    //    output(myPara.oErrName);
    //    exit(1);
    //}
    ipair_t hdUb(myPara.hdMax, myPara.hdMax);

    ipair_t dPoints(0, myPara.K - myPara.step); //dividing points //a
    
    std::string myRead = addr;
    int readLen = myRead.length();
    int nextPos = 0, prePos = -1;
    bool crawlingFlag = false;
    int crawlTo = -1;

    while (1) {

        upair_t mosaic;
        uint64_t ID1, ID2;
        if (!(toID(ID1, &myRead[nextPos], myPara.K) &&
                toID(ID2, &myRead[nextPos + myPara.step], myPara.K)))
            return; /* containing non-acgt*/
        mosaic = upair_t(ID1, ID2);

        if (errorCall(mosaic, dPoints, hdUb, qAddr + nextPos, myPara)) {//a
        //if (errorCall(mosaic, hdUb, qAddr + nextPos, myPara)) {

            updateRead(myRead, nextPos);

            //std::cout << myRead << "\n"; //debug

            if (nextPos + myPara.step + myPara.K == readLen) return;

            prePos = nextPos;

            if (!crawlingFlag) {
                if (nextPos + myPara.step <= readLen - myPara.K - myPara.step) {
                    dPoints = ipair_t(myPara.K, 0);
                    nextPos += myPara.step;
                    hdUb = ipair_t(0, myPara.hdMax);
                    //mosaic.first = mosaic.second;
                    //if (!toID(mosaic.second, addr + nextPos + myPara.step, myPara.K))
                    //    return;
                } else {
                    dPoints = ipair_t(myPara.K, nextPos + myPara.K + 2 * myPara.step - readLen);
                    nextPos = readLen - myPara.K - myPara.step;
                    hdUb = ipair_t(0, myPara.hdMax);
                }

              //  std::cout << "next_pos = " << nextPos << "\n";
            }
        } 
        else {
            if (crawlingFlag == false){
                crawlingFlag = true;
                crawlTo = nextPos;
            }
        }

        if (crawlingFlag){

         // crawling --- to do reverse crawling
            if (nextPos == prePos + 1 || prePos == -1)  return;

            //if (crawlingFlag == false) {
            //    crawlTo = nextPos;
            //    crawlingFlag = true;
            // }
            nextPos = prePos + 1;

            if(nextPos == crawlTo) crawlingFlag = false;
          //  std::cout << "next_pos = " << nextPos << "\n";

            //hdUb = ipair_t(0, myPara.hdMax);
            hdUb = ipair_t(0, 1);
            dPoints = ipair_t(myPara.K, myPara.K - 3);

        }
    }
}

void Parser::updateRead(std::string& myRead, int shift) {

    if (readErr_.size()) {        
        // 1. adding shift to err positions w.r.t read & update read
        for (int i = 0; i < readErr_.size(); ++i) {
            readErr_[i].pos += shift;
            myRead[readErr_[i].pos] = bits_to_char(readErr_[i].from);
        }
    //    std::cout << "ec !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"; //debug
    //    std::cout << myRead << "\n";
        // 2. update records_
        if (records_.size() && records_.back().readID == readID_) {
            (records_.back()).evec.insert((records_.back()).evec.end(),
                    readErr_.begin(), readErr_.end());
        } else {
            records_.push_back(record_t(readID_, readErr_));
        }

        // 3. update snKArray_ and lnKArray_ --- todo

        readErr_.clear();
    }
}

bool increKC(kcvec_t& kcvec, uint64_t ID, const kcvec_t& KCArray) {

    int idx = bisearch(KCArray, KCArray.size(), ID);
    if (idx != -1) {
        kcvec.push_back(KCArray[idx]);
        return true;
    }
    return false;
}

void candidates(ivec_t& candies, const kcvec_t& tiles, int threshold) {
    for (int i = 1; i < tiles.size(); ++i) {
        if (tiles[i].goodCnt >= threshold)
            candies.push_back(i);
    }
}

/*
int candidates(ivec_t& candies, const kcvec_t& tiles, int threshold) {
    int max = -1, idx;
    for (int i = 0; i < tiles.size(); ++i) {
        if (tiles[i].goodCnt > max) {
            max = tiles[i].goodCnt;
            idx = i;
        }
        if (i != 0 && tiles[i].goodCnt >= threshold)
            candies.push_back(i);
    }
    return idx;
}
 * */

void diff(evec_t& errs, uint64_t from, uint64_t to, char* qAddr, int len) {
    uint64_t tmp = from ^ to;
    int idx = 0;
    while (tmp) {
        if (tmp & 0x3) {
            e_t err;
            err.pos = len - 1 - idx;
            err.from = from & 0x3;
            err.to = to & 0x3;
            err.qual = (int) qAddr[(len - 1) - idx];
            errs.push_back(err);
        }
        tmp >>= 2;
        from >>= 2;
        to >>= 2;
        idx++;
    }
}

bool goodQuals(char* qAddr, int len, int threshold) {
    for (int i = 0; i < len; ++i) {
        if (qAddr[i] < threshold)
            return false;
    }
    return true;
}


bool Parser::errorCall(const upair_t& mosaic, const ipair_t& dPoints,
        const ipair_t& hdUb, char* qAddr, const Para& myPara) {

//bool Parser::errorCall(const upair_t& mosaic, const ipair_t& hdUb,
//        char* qAddr, const Para& myPara) {

    upair_t rvMosaic(
            reverse_complementary<uint32_t, uint32_t > (mosaic.first, myPara.K),
            reverse_complementary<uint32_t, uint32_t > (mosaic.second, myPara.K));

    kcvec_t heads, tails, rvHeads, rvTails;

    increKC(heads, mosaic.first, snKArray_);
    increKC(tails, mosaic.second, snKArray_);

    /*
    if (!increKC(heads, mosaic.first, snKArray_)) {
        std::cout << "initHeads err: " << mosaic.first << " not exist\n";
        exit(1);
    }
    if (!increKC(tails, mosaic.second, snKArray_)) {
        std::cout << "initTails err: " << mosaic.second << " not exist\n";
        exit(1);
    }
     */
    increKC(rvHeads, rvMosaic.second, snKArray_);
    increKC(rvTails, rvMosaic.first, snKArray_);
    
    ipair_t hd(0, 0);
    while (hd.first <= hdUb.first) {

        kcvec_t tiles, rvTiles;
        tiling(tiles, heads, tails, myPara);
  //      print_kcvec("tiles", tiles, myPara.K + myPara.step);
        tiling(rvTiles, rvHeads, rvTails, myPara);
  //      print_kcvec("rvtiles", rvTiles, myPara.K + myPara.step);
        mergeTiles(tiles, rvTiles, myPara);
  //      print_kcvec("After Merging", tiles, myPara.K + myPara.step);

        /*
         * Error Calling Current Tile
         * tiles.size() may equal to 0 due to error correction
         */
        if (tiles.size() > 0) {
            if (tiles[0].goodCnt >= myPara.tGoodTile) return true;
            else {
                if (tiles.size() == 1) { // try to correct
                    if (tiles[0].goodCnt >= myPara.tCard &&
                            goodQuals(qAddr, myPara.K + myPara.step, myPara.Qlb))
                        return true;
                } else {
                    ivec_t candies; // indices of tiles
                    candidates(candies, tiles, myPara.tCard);

                    if (tiles[0].goodCnt >= myPara.tCard) {
                        /*
                         * ec only if non-ambig correction of low qual bases
                         * could be identified and correcting to one of the
                         * high cardinality neighbors
                         */
                        int tGoodCnt = tiles[0].goodCnt / myPara.tRatio;

                        ivec_t highCardNbs;
                        for (int i = 0; i < candies.size(); ++i) {
                            if (tiles[candies[i]].goodCnt >= tGoodCnt)
                                highCardNbs.push_back(candies[i]);
                        }
                        int alterNum = 0;
                        for (int i = 0; i < highCardNbs.size(); ++i) {
                            /*
                             * check all candidates, if ambig, then do not ec
                             */
                            evec_t errs;
                            diff(errs, tiles[highCardNbs[i]].ID, tiles[0].ID,
                                    qAddr, myPara.step + myPara.K);
                            for (int j = 0; j < errs.size(); ++j) {
                                if (errs[j].qual < myPara.Qlb) {
                                    readErr_ = errs;
                                    alterNum++;
                                    break;
                                }
                            }
                        }
                        if (alterNum > 1) { // ambig
                            readErr_.clear();
                        } else return true;
                    } else {
                        /*
                         * ec only if non-ambig correction to one of [candies]
                         * could be identified
                         */
                        if (candies.size() == 1) {
                            diff(readErr_, tiles[candies[0]].ID, tiles[0].ID,
                                    qAddr, myPara.step + myPara.K);
                            return true;
                        }
                    }
                }
            }
        }
        /*
         * increase Hamming Distance to search for more neighbors
         */
        hd.second++;
        if (hd.second <= hdUb.second) {
            unitNeighbor(tails, dPoints.second, myPara);
            unitNeighbor(rvHeads, dPoints.second, myPara);
            //unitNeighbor(tails, myPara);
            //unitNeighbor(rvHeads, myPara);
        } else {
            hd.first++;
            if (hd.first <= hdUb.first) {
                unitNeighbor(heads, dPoints.first, myPara);
                unitNeighbor(rvTails, dPoints.first, myPara);
                //unitNeighbor(heads, myPara);
                //unitNeighbor(rvTails, myPara);
            }
            else{ // all possible searches have been done, till this stage
                  // if tile[0] is the maximum and > tCard then, it is considered
                  // as correct due to low coverage region
                int maxCnt = 0;
                bool tflag = true;
                if (tiles.size() > 0) maxCnt = tiles[0].goodCnt;
                if (maxCnt >= myPara.tCard) {
                    for (int i = 1; i < tiles.size(); i ++){
                        if (maxCnt < tiles[i].goodCnt) {
                            tflag = false;
                            break;
                        }
                    }
                    if (tflag) return true;
                }                
            }
        }
    }    
    return false;
}

bool checkPoint(const ivec_t& indices, int dPoint) {
    for (int i = 0; i < indices.size(); ++i) {
        if (indices[i] >= dPoint)
            return true;
    }
    return false;
}

void Parser::unitNeighbor(kcvec_t& myNB, int dPoint, const Para& myPara) {

//void Parser::unitNeighbor(kcvec_t& myNB, const Para& myPara) {

    if (myNB.size() == 0) return;

    uvec_t inputIDs(myNB.size(), 0);
    for (int i = 0; i < myNB.size(); inputIDs.push_back(myNB[i].ID), ++ i);
    std::sort(inputIDs.begin(), inputIDs.end());
    
    uvec_t nbIDs; // neigbhoring IDs

    for (int i = 0; i < myNB.size(); ++i) {

        uint32_t myID = myNB[i].ID;

        uvec_t myNeighbor; //store ID's neighbors by searching in all tables

        /*
         *  search in every table
         */

        int tableSize = table_.size() / masks_.size();
        for (int j = 0; j < masks_.size(); ++j) {

            if (!checkPoint(maskIdx_[j], dPoint)) continue; //a

            std::pair<uvec_t::iterator, uvec_t::iterator> bounds;

            bounds = std::equal_range(table_.begin() + j*tableSize,
                    table_.begin() + (j + 1) * tableSize, myID, TComp(masks_[j]));

            // identify if HD = 1
            //uint32_t rMask = ~masks_[j];
            for (uvec_t::iterator it = bounds.first; it != bounds.second; ++it) {
                // check if differed position satisfies dPoint
                int idx = myPara.K - 1;
                bool dPointFlag = true;

                //if (*it == myID) continue;
                if (std::binary_search(inputIDs.begin(), inputIDs.end(), *it)) continue;
                
                //uint32_t tmp = ((*it) & rMask) ^ (myID & rMask);
                uint32_t tmp = (*it) ^ myID;
                int cnt = 0;
                while (tmp) {
                    
                    if (idx < dPoint) {
                        dPointFlag = false;
                        break;
                    }
                    
                    if (tmp & 0x3) cnt++;
                    tmp >>= 2;
                    //idx--;
                }
                if (cnt == 1 && dPointFlag) {
                //if (cnt == 1) {
                    myNeighbor.push_back(*it);
                }
            }
        }

        nbIDs.insert(nbIDs.begin(), myNeighbor.begin(), myNeighbor.end());
    }

    /*
     * remove duplicates
     */
    std::sort(nbIDs.begin(), nbIDs.end());
    uvec_t::iterator it
            = std::unique_copy(nbIDs.begin(), nbIDs.end(), nbIDs.begin());
    nbIDs.resize(it - nbIDs.begin());

    /*
     * now have all nbIDs from table, then search in [snKArray_]
     * to get complete multiplicity and quality score information
     */

    for (int i = 0; i < nbIDs.size(); ++i) {

        int idx = bisearch(snKArray_, snKArray_.size(), nbIDs[i]);
        if (idx != -1){
            if(snKArray_[idx].goodCnt >= myPara.tCard) {
                myNB.push_back(snKArray_[idx]);
            }
        }
    }
}

/*
 * At current stage, only 1-neighbor is considered. Need be extended
 * to derive HD > 1 neighbors. So for a particular region of the read,
 * we could consider more neighbors to accomendate qual scores
 */
/*
void Parser::neighbors(std::vector<kcvec_t>& nodes,
        const uvec_t& repIDs, const Para& myPara) {


    
       //deal with each ID for a read
     

    for (int i = 0; i < repIDs.size(); ++i) {

        uint32_t ID = repIDs[i];
        uvec_t IDArray; //store ID and its neighbors by searching in all tables

        IDArray.push_back(ID);
        
         //  search in every table
         

        int tableSize = table_.size() / masks_.size();
        for (int j = 0; j < masks_.size(); ++j) {

            std::pair<uvec_t::iterator, uvec_t::iterator> bounds;

            bounds = std::equal_range(table_.begin() + j*tableSize,
                    table_.begin() + (j + 1) * tableSize, ID, TComp(masks_[j]));

            // identify if HD = 1
            uint32_t rMask = ~masks_[j];
            for (uvec_t::iterator it = bounds.first; it != bounds.second; ++it) {
                if (*it == ID) continue;
                uint32_t tmp = ((*it) & rMask) ^ (ID & rMask);
                int cnt = 0;
                while (tmp) {
                    if (tmp & 0x3) cnt++;
                    tmp >>= 2;
                }
                if (cnt == 1) {
                    IDArray.push_back(*it);
                }
            }

            //IDArray.insert(IDArray.end(), bounds.first, bounds.second);
        }
        
         // remove duplicates
         
        std::sort(IDArray.begin(), IDArray.end());
        uvec_t::iterator it
                = std::unique_copy(IDArray.begin(), IDArray.end(), IDArray.begin());
        IDArray.resize(it - IDArray.begin());

        
        // now have all IDs from table, then search in [snKArray_]
        // to get complete multiplicity and quality score information
         
        for (int j = 0; j < IDArray.size(); ++j) {
            int idx = bisearch(snKArray_, snKArray_.size(), IDArray[j]);
            if (idx == -1) {
                std::cout << "err: neighbors bisearch found nothing !\n";
                exit(1);
            }
            if (IDArray[j] == ID && nodes[i].size() > 0) {//insert to front
                nodes[i].push_back(nodes[i][0]);
                nodes[i][0] = snKArray_[idx];
            } else {
                nodes[i].push_back(snKArray_[idx]);
            }
        }

        // debugging print out all IDs as strings
        //print_kcvec("", nodes[i], myPara.K);
    }
}
*/
/*
 * Keep elements of N only if the ID equals to the last 2*len bits in tiles
 */
void updateNodes(kcvec_t& N, kcvec_t& N_rv, const kcvec_t& tiles, int len) {

    std::set<uint64_t> IDs, IDs_rv;
    kcvec_t tmpVec;
    for (int i = 0; i < tiles.size(); ++i) {
        uint64_t last2k = tiles[i].ID & ((0x1 << 2 * len) - 1);
        IDs.insert(last2k);
        IDs_rv.insert(reverse_complementary <uint32_t, uint32_t > (last2k, len));
    }
    for (int i = 0; i < N.size(); ++i) {
        if (IDs.count(N[i].ID)) tmpVec.push_back(N[i]);
    }
    N = tmpVec;
    tmpVec.clear();
    for (int i = 0; i < N_rv.size(); ++i) {
        if (IDs_rv.count(N_rv[i].ID)) tmpVec.push_back(N_rv[i]);
    }
    N_rv = tmpVec;
}

/*
void Parser::errCall(std::vector<kcvec_t>& nodes, std::vector<kcvec_t>& nodes_rv,
        const std::string& qual, const Para& myPara) {

    int repNum = nodes.size();

    bvec_t track(qual.length(), false); // tracking the corrected positions

    int p = 0;
    while (p <= repNum - 2) { // consider p & p+1

        kcvec_t tiles, tiles_rv;

        tiling(tiles, nodes[p], nodes[p + 1], myPara);
        tiling(tiles_rv, nodes_rv[p + 1], nodes_rv[p], myPara);
        // merge [tiles_rv] to [tiles]
        mergeTiles(tiles, tiles_rv, myPara);

        
         // Err Calling Rules
         
        if (tiles.size() == 1) {
            if (tiles[0].goodCnt >= myPara.tCard) { // pass the test
                //set validated pos as true
                for (int i = p * myPara.step; i < p * myPara.step + myPara.step
                        + myPara.K; track[i] = true, i++);
            } else {
                //re_search(track, 
            }
        } else {

        }
        //according to Err Calling rslt, modify nodes[p+1] and nodes_rv[p+1]
         // to be the last 2K bits of tiles & reverse_compl
        //  But no need to update the last nodes
        
        if (p != repNum - 2) {
            updateNodes(nodes[p + 1], nodes_rv[p + 1], tiles, myPara.K);
        }
        std::cout << "p = " << p << "\n";
        print_kcvec("Tiles", tiles, myPara.K + myPara.step);

        p++;
    }
}
*/

bool unicpy(const kc_t& e1, const kc_t& e2) {
    return (e1.ID == e2.ID);
}

void Parser::mergeTiles(kcvec_t& tileTo, kcvec_t& tileFrom, const Para& myPara) {
    /*
     *  convert IDs in [tileFrom] to revcompl IDs and insert to [tileTo]
     *  then sort and merge (remove duplicates); keep the repID in place
     */
    if (tileFrom.size() == 0) return;

    //print_kcvec("TileTo", tileTo, myPara.K + myPara.step);
    //print_kcvec("TileFrom", tileTo, myPara.K + myPara.step);

    /* convert to reverse compl IDs
     */
    for (int i = 0; i < tileFrom.size(); ++i) {
        tileFrom[i].ID = reverse_complementary <uint64_t, uint64_t >
                (tileFrom[i].ID, myPara.K + myPara.step);

    }

    if (tileTo.size() == 0){
        tileTo = tileFrom;
        tileFrom.clear();
        return;
    }

    uint64_t repID = tileTo[0].ID;
    tileTo.insert(tileTo.end(), tileFrom.begin(), tileFrom.end());
    tileFrom.clear();

    /* sort
     */
    std::sort(tileTo.begin(), tileTo.end(), Knumcomp());

    /* merge
     */
    //print_kcvec("Before Merging", tileTo, myPara.K + myPara.step);
    int idx1 = 0;

    for (int idx2 = idx1 + 1; idx2 < tileTo.size(); ++idx2) {
        if (tileTo[idx2].ID == tileTo[idx1].ID) {
            tileTo[idx1].goodCnt += tileTo[idx2].goodCnt;
            tileTo[idx1].cnt += tileTo[idx2].cnt;
        } else {
            idx1 = idx2;
        }
    }

    kcvec_t::iterator it =
            std::unique_copy(tileTo.begin(), tileTo.end(), tileTo.begin(), unicpy);
    tileTo.resize(it - tileTo.begin());

    /* keep repID in the first pos
     */
    if (tileTo[0].ID != repID) {
        for (int i = 1; i < tileTo.size(); ++i) {
            if (tileTo[i].ID == repID) {
                kc_t tmp = tileTo[0];
                tileTo[0] = tileTo[i];
                tileTo[i] = tmp;
                break;
            }
        }
    }

    //print_kcvec("After Merging", tileTo, myPara.K + myPara.step);
    //exit(1);
}

/*
 * flag: true -- forward    flase--reverse 
 */
void Parser::tiling(kcvec_t& tiles, const kcvec_t& N1,
        const kcvec_t& N2, const Para& myPara) {

    if (N1.size() == 0 || N2.size() == 0) return;
    uint64_t reptile;
    if (!overlay(reptile, N1[0].ID, N2[0].ID, myPara)) {
        std::cout << "Err: errCall, reptile construction fail\n";
        exit(1);
    }
    int idx = bisearch(lnKArray_, lnKArray_.size(), reptile);
    if (idx != -1) {
        tiles.push_back(lnKArray_[idx]);
    }
    else {
        tiles.push_back(kc_t(reptile, 0, 0));
    }
    for (int i = 0; i < N1.size(); ++i) {
        for (int j = 0; j < N2.size(); ++j) {
            if (i == 0 && j == 0) continue;
            uint64_t tmptile;
            if (overlay(tmptile, N1[i].ID, N2[j].ID, myPara)) {
                //binary search 
                idx = bisearch(lnKArray_, lnKArray_.size(), tmptile);

                if (idx != -1) {
                    tiles.push_back(lnKArray_[idx]);
                }
            }
        }
    }
    //debug code
    //print_kcvec("", tiles, myPara.K + myPara.step);    
}

bool Parser::overlay(uint64_t& rslt, uint64_t n1, uint64_t n2, const Para& myPara) {

    int shift = 2 * myPara.step;
    // comparison suffix - prefix if necessary
    if (myPara.step >= myPara.K || 
            (n2 >> shift) == (n1 & ((1 << 2*(myPara.K - myPara.step)) - 1))) {
        rslt = (n1 << shift) | n2;
        return true;
    }
    return false;
}

void Parser::output(const std::string& filename) {

    /*
     * Resulting file format:
     * ReadID (sorted)  ErrNum  [pos from to qual] [pos from to qual]...
     * from: reference; to: read (numercial value Aa:0 Cc:1 Gg:2 Tt:3 others 4)
     * qual: quality (numerical value)
     */
    std::ofstream oHandle(filename.c_str());
    if (!oHandle.good()) {
        std::cout << "open " << filename << " failed, correct path?\n";
        exit(1);
    }

    for (int i = 0; i < records_.size(); ++i) {

        oHandle << records_[i].readID << "\t" << records_[i].evec.size();

        for (int j = 0; j < records_[i].evec.size(); ++j) {
            oHandle << "\t" << records_[i].evec[j].pos << "\t"
                    << records_[i].evec[j].from << "\t" << records_[i].evec[j].to
                    << "\t" << records_[i].evec[j].qual;
        }
        oHandle << "\n";
    }
    oHandle.close();
}

/*
void Parser::merge(const Kmer& sW, const Kmer& lW, const Para& myPara) {

    int incre = sW.getElem().size();
    std::cout << "\norgi size = " << sWindow_.size() << '\n';
    std::cout << "incre = " << incre << "\n";
    int oriSize = (int) sWindow_.size();
    sWindow_.reserve(oriSize + incre);
    int trueIncre = 0;
    for (int i = 0; i < incre; ++i) {
        kc_s tmp = sW.getElem()[i];
        size_t ID;
        if (toID(ID, tmp.repaddr, myPara.K)) {

            int idx = bisearch(sWindow_, oriSize, ID);

            if (idx == -1) {
                sWindow_.push_back(kc_t(ID, tmp.goodCnt, tmp.cnt));
                trueIncre++;
            } else {
                sWindow_[idx].cnt += tmp.cnt;
                sWindow_[idx].goodCnt += tmp.goodCnt;
            }
        }
    }
    std::cout << "true incre = " << trueIncre << "\n";
    sWindow_.resize(oriSize + trueIncre);
    std::cout << "updated size: " << sWindow_.size() << "\n";
    std::sort(sWindow_.begin(), sWindow_.end(), Knumcomp());
}

void Parser::printKID2Cnt(const Para& myPara){

    std::cout << "output to " << myPara.oKCntName << "\t";

    std::ofstream output(myPara.oKCntName.c_str());

    if(!output.good()){
        std::cout << "cant' write to file: " << myPara.oKCntName << std::endl;
        exit(1);
    }

    std::map<size_t, ipair_t>::iterator it (lWindow_.begin());
    for (; it != lWindow_.end(); ++ it){

        output << toString((*it).first, myPara.K + myPara.step) << "\t";
        output << (*it).second.second << "\n";
    }

    output.close();
    std::cout << "done\n";
}
 */
